#ifndef POLYDIVIDE_H
#define POLYDIVIDE_H

#include <math.h>
#include <stdlib.h>
#include <string.h>

// Takes two polynomials a, b, ordered in order of increasing power, terminated
// by INFINITY and a pointer to a pre-allocated remainder polynomial r.
// Allocates memory to q and r
double *entry(double const *a, double const *b, double **r) {
  // We first need to find the end of the polynomial
  // This is because polynomial division starts at the highest power
  // This also lets us compute the length of the polynomials,
  // important because we to compute the length of the quotient
  // to allocate memory for it.
  double const *a_end = a, *b_end = b;

  while (!isinf(*a_end)) { ++a_end; }
  while (!isinf(*b_end)) { ++b_end; }

  // The number of coeffients (excludes INFINITY terminator)
  // This is one more than the degree of the polynomial
  size_t const a_len = a_end - a;
  size_t const b_len = b_end - b;

  // Handle the numerator having lower degree than denominator
  if (a_len < b_len) {
    double *const q = malloc(1 * sizeof *q);
    *q = INFINITY;

    // Just create a copy of the numerator as the remainder
    *r = malloc((a_len + 1) * sizeof **r);
    memcpy(*r, a, (a_len + 1) * sizeof **r);

    return q;
  }

  size_t const q_len = a_len - b_len + 1;

  double *const q = malloc((q_len + 1) * sizeof *q);
  double *q_end = q + q_len;

  // Terminate the quotient polynomial
  *q_end = INFINITY;
  --q_end;

  // Move back to the coefficient before INFINITY
  --a_end;
  --b_end;

  // temporary values for computation
  double *temps = malloc(b_len * sizeof *temps);
  double *const temps_end = temps + b_len - 1;
  memcpy(temps, a_end - b_len + 1, b_len * sizeof *temps);

  // a_end - a + 1 = number of coefficients left in a
  while (a_end - a + 1 >= b_len) {
    double const a_lead = *temps_end;
    double const b_lead = *b_end;

    double const lead_factor = a_lead / b_lead;

    *q_end = lead_factor;

    // Recalculate temporaries
    // Basically subtracting b * lead_factor from temps,
    // shifting over by one,
    // and then getting the next coefficient from a
    double *tptr;
    double const *bptr;
    for (tptr = temps_end, bptr = b_end; tptr != temps; --tptr, --bptr) {
      *tptr = *(tptr - 1) - lead_factor * *(bptr - 1);
    }

    // Last iteration
    // Since it's the last iteration, there are no more coefficients in a
    if (a_end - a + 1 == b_len) {
      break;
    }

    *temps = *(a_end - b_len);

    --a_end;
    --q_end;
  }

  // Temps now basically contains the remainder, just needs some touching up
  // meaning shifting it over by one and adding the INFINITY terminator
  memmove(temps, temps + 1, (b_len - 1) * sizeof *temps);
  *temps_end = INFINITY;

  *r = temps;
  return q;
}

#endif
